/** \file src/xtal/xtal_iterator.h
* \brief src/xtal/xtal_iterator.h
*/

#ifndef XTAL_ITERATOR_H_INCLUDE_GUARD
#define XTAL_ITERATOR_H_INCLUDE_GUARD

#pragma once

namespace xtal{

class DelegateToIterator : public Base{
	IDPtr member_;
public:
	
	DelegateToIterator(const IDPtr& name)
		:member_(name){}

	void on_rawcall(const VMachinePtr& vm);
};

/**
* \xbind lib::builtin
* \brief Ce[^֗̕\bh񋟂邽߂̃NX
*/
class Iterator{
public:

	//void scan(pattern);

	/**
	* \xbind
	* \brief vfnfobOvg
	* ̃Ce[^ԂB
	*/
	AnyPtr ip(int_t n = 3);

	/**
	* \xbind
	* \brief vfZ߂ĔzɂĕԂ
	*/
	ArrayPtr op_to_array();

	/**
	* \xbind
	* \brief vfZ߂ĘAzzɂĕԂ
	* ȏ̒l񋓂Ce[^łKv
	*/
	MapPtr op_to_map();

	/**
	* \xbind
	* \brief vf𕶎񉻂AɂsepŘAԂ
	* \code
	* [1, 2, 3].join("::").p;
	* //=> 1::2::3
	* \endcode
	*/
	StringPtr join(const StringPtr& sep = "");

	/**
	* \xbind
	* \brief (index, value)ԂCe[^Ԃ
	* [vƂindex̓CNgB
	* \code
	* [45, 12, 33].with_index{ |index, value|
	*     %f(%d-%d)(index, value).p;
	* }
	* //=> 0-45
	* //=> 1-12
	* //=> 2-33
	* \endcode
	*/
	AnyPtr with_index(int_t start = 0);

	/**
	* \xbind
	* \brief vfconv֐ŕϊvfԂCe[^Ԃ
	* map\bhƓӖł
	*/
	AnyPtr collect(const AnyPtr& conv);

	/**
	* \xbind
	* \brief vfconv֐ŕϊvfԂCe[^Ԃ
	* \code
	* [10, 20, 30].map(|x| x*2){
    *	it.p;
	* }
	* //=> 20
	* //=> 40
	* //=> 60
	* \endcode
	*/
	AnyPtr map(const AnyPtr& conv);

	/**
	* \xbind
	* \brief IʂꂽvfԂCe[^Ԃ
	* filter\bhƓ̈Ӗł
	*/
	AnyPtr select(const AnyPtr& pred);

	/**
	* \xbind
	* \brief IʂꂽvfԂCe[^Ԃ
	* \code
	* [1, 2, 3].filter(|x| x%2==1){
	*	it.p;
	* }
	* //=> 1
	* //=> 3
	* \endcode
	*/
	AnyPtr filter(const AnyPtr& pred);

	/**
	* \xbind
	* \brief pred֐trueԂ܂ŃCe[gCe[^Ԃ
	* \code
	* [7, 5, 3, 2, 1, 3].break_if(|x| x%2!=1){
	*     it.p;
	* }
	* //=> 7
	* //=> 5
	* //=> 3
	* \endcode
	*/
	AnyPtr break_if(const AnyPtr& pred);

	/**
	* \xbind
	* \brief timesCe[gCe[^Ԃ
	* \code
	* [5, 3, 7].cycle.take(5).join(",").p;
	* //=> 5,3,7,5,3
	* \endcode
	*/
	AnyPtr take(int_t times);

	/**
	* \xbind
	* \brief ɃCe[gCe[^Ԃ
	*/
	AnyPtr zip(...);

	/**
	* \xbind
	* \brief ɃCe[gCe[^Ԃ
	*/
	AnyPtr chain(...);

	/**
	* \xbind
	* \brief Ce[gvfȂȂAŏJԂCe[^Ԃ
	* \code
	* [4, 8, 11].map(|x| x*3).cycle{
	*    it.p;
	* }
	* //=> 12
	* //=> 24
	* //=> 33
	* //=> 12
	* //=> 24
	* //=> 33
	* // ȉJԂ
	* \endcode
	*/
	AnyPtr cycle();

	/**
	* \xbind
	* \brief Ce[gvf̍ő̒lԂ
	* \param pred vf肷q֐IuWFNg
	* ̈󂯎Aڂڂ̕傫ȂtrueԂ֐nƁB
	* nullw肷 < ZqŔ肳B
	*/
	AnyPtr max_element(const AnyPtr& pred = null);

	/**
	* \xbind
	* \brief Ce[gvf̍ŏ̒lԂ
	* \param pred vf肷q֐IuWFNg
	* ̈󂯎Aڂڂ̕ȂtrueԂ֐nƁB
	* nullw肷 > ZqŔ肳B
	*/
	AnyPtr min_element(const AnyPtr& pred = null);

	/**
	* \xbind
	* \brief Ce[gvfpred֐trueԂ֐Ԃ
	* \param pred TĂvf𔻒肷q֐IuWFNg
	* ̈󂯎ATĂvfȂtrueԂ֐nƁB
	*/
	AnyPtr find(const AnyPtr& pred);

	/**
	* \xbind
	* \brief inject
	*/
	AnyPtr inject(const AnyPtr& init, const AnyPtr& fn);

	/**
	* \xbind
	* \brief (O̗vf, ̗vf)vfCe[^Ԃ
	* ŏ̈̑O̗vfɂundefined
	*/
	AnyPtr with_prev();

};

/**
* \xbind lib::builtin
* \brief Ce[g\ȃReî߂̎񋟂邽߂̃NXB
* inheritNX̃IuWFNǵAeach\bhoRlib::builtin::Iterator\bhĂяoB
*/
class Iterable{};

class ZipIter : public Base{
public:

	ZipIter(const VMachinePtr& vm);

	void common(const VMachinePtr& vm, const IDPtr& id);

	void block_first(const VMachinePtr& vm);
	
	void block_next(const VMachinePtr& vm);

	void block_break(const VMachinePtr& vm);

	void on_visit_members(Visitor& m);

	ArrayPtr next_;
};

struct BlockValueHolder1{
	
	BlockValueHolder1(const AnyPtr& tar, bool& not_end);
	BlockValueHolder1(const ArrayPtr& tar, bool& not_end);
	~BlockValueHolder1();

	AnyPtr target;
	AnyPtr values[1];

	SmartPtr<ArrayIter> it;
};

struct BlockValueHolder2{
	
	BlockValueHolder2(const AnyPtr& tar, bool& not_end);
	BlockValueHolder2(const MapPtr& tar, bool& not_end);
	~BlockValueHolder2();

	AnyPtr target;
	AnyPtr values[2];

	SmartPtr<MapIter> it;
};

struct BlockValueHolder3{
	BlockValueHolder3(const AnyPtr& tar, bool& not_end);
	~BlockValueHolder3();

	AnyPtr target;
	AnyPtr values[3];
};

bool block_next(BlockValueHolder1& holder, bool first);
bool block_next(BlockValueHolder2& holder, bool first);
bool block_next(BlockValueHolder3& holder, bool first);
void block_break(AnyPtr& target);

}

#endif // XTAL_ITERATOR_H_INCLUDE_GUARD
